#include "atomicOperations.h"
#include "isamax_macros.h"
#include <cfloat>
// 
// Generic vector max kernel.
//
// - Designed for architectures with atomicCAS, i.e., sm_11 and higher.
// - Simple design for all capabilities
// - Handled by splitting matrix as
// --------------------------
// |                 |      |
// |                 |      |
// |                 |      |
// | left tiled part | rest | 
// |                 |      |
// |                 |      |
// |                 |      |
// --------------------------
// - Kernel is generic and works for all cases
// - Unrolling of inner loops
//     1) to reduce instruction count from loop
//     2) to collect shared loads into 128 bit transactions (compiler should do that)
//   * Level determined by WORKSIZE_n
//   * For WORKSIZE_n = 1:
//       Given by a constant: full unroll
//       Given by a variable: unroll level 4

#define imax(a,b,c,d) if (a < b) { a = b; c = d; }

template <int BLOCKSIZE_n>
inline __device__ void block_reduce_max(const float *x, int *y, volatile float *smem, volatile int *imem, float sum, int index, int tid)
{
	smem[tid] = sum;
	imem[tid] = index;
	__syncthreads();
	if (BLOCKSIZE_n >= 1024) { if (tid < 512) { imax(smem[tid], smem[tid + 512], imem[tid], imem[tid + 512]); } __syncthreads(); }
	if (BLOCKSIZE_n >= 512) { if (tid < 256) { imax(smem[tid], smem[tid + 256], imem[tid], imem[tid + 256]); } __syncthreads(); }
	if (BLOCKSIZE_n == 480) { if (tid < 224) { imax(smem[tid], smem[tid + 256], imem[tid], imem[tid + 256]); } __syncthreads(); }
	if (BLOCKSIZE_n == 448) { if (tid < 192) { imax(smem[tid], smem[tid + 256], imem[tid], imem[tid + 256]); } __syncthreads(); }
	if (BLOCKSIZE_n == 416) { if (tid < 160) { imax(smem[tid], smem[tid + 256], imem[tid], imem[tid + 256]); } __syncthreads(); }
	if (BLOCKSIZE_n == 384) { if (tid < 128) { imax(smem[tid], smem[tid + 256], imem[tid], imem[tid + 256]); } __syncthreads(); }
	if (BLOCKSIZE_n >= 256) { if (tid < 128) { imax(smem[tid], smem[tid + 128], imem[tid], imem[tid + 128]); } __syncthreads(); }
	if (BLOCKSIZE_n == 352) { if (tid < 96) { imax(smem[tid], smem[tid + 256], imem[tid], imem[tid + 256]); } __syncthreads(); }
	if (BLOCKSIZE_n == 224) { if (tid < 96) { imax(smem[tid], smem[tid + 128], imem[tid], imem[tid + 128]); } __syncthreads(); }
	if (BLOCKSIZE_n == 320) { if (tid < 64) { imax(smem[tid], smem[tid + 256], imem[tid], imem[tid + 256]); } __syncthreads(); }
	if (BLOCKSIZE_n == 192) { if (tid < 64) { imax(smem[tid], smem[tid + 128], imem[tid], imem[tid + 128]); } __syncthreads(); }
	if (BLOCKSIZE_n >= 128) { if (tid < 64) { imax(smem[tid], smem[tid + 64], imem[tid], imem[tid + 64]); } __syncthreads(); }
	if (BLOCKSIZE_n == 288) { if (tid < 32) { imax(smem[tid], smem[tid + 256], imem[tid], imem[tid + 256]); } __syncthreads(); }
	if (BLOCKSIZE_n == 160) { if (tid < 32) { imax(smem[tid], smem[tid + 128], imem[tid], imem[tid + 128]); } __syncthreads(); }
	if (BLOCKSIZE_n == 96) { if (tid < 32) { imax(smem[tid], smem[tid + 64], imem[tid], imem[tid + 64]); } __syncthreads(); }
	if (BLOCKSIZE_n >= 64) { if (tid < 32) { imax(smem[tid], smem[tid + 32], imem[tid], imem[tid + 32]); } }
	if (BLOCKSIZE_n >= 32) { if (tid < 16) { imax(smem[tid], smem[tid + 16], imem[tid], imem[tid + 16]); } }
	if (BLOCKSIZE_n >= 16) { if (tid < 8) { imax(smem[tid], smem[tid + 8], imem[tid], imem[tid + 8]); } }
	if (BLOCKSIZE_n >= 8) { if (tid < 4) { imax(smem[tid], smem[tid + 4], imem[tid], imem[tid + 4]); } }
	if (BLOCKSIZE_n >= 4) { if (tid < 2) { imax(smem[tid], smem[tid + 2], imem[tid], imem[tid + 2]); } }
	if (BLOCKSIZE_n >= 2) { if (tid < 1) { imax(smem[tid], smem[tid + 1], imem[tid], imem[tid + 1]); } }
	if (tid == 0) { atomicImax(x, smem[tid], y, imem[tid]); }
}

template <int BLOCKSIZE_n, int WORKSIZE_m, int WORKSIZE_n>
__launch_bounds__(BLOCKSIZE_n, MIN_BLOCKS_PER_MP)
__global__ void isamax_kernel(const float *x, int *y, int n, int ntile)
{
	const int tid = threadIdx.x;
	const int idx = blockIdx.x * BLOCKSIZE_n * WORKSIZE_n;
	int index = idx + tid, index_i = index;
	const float *x_i = x + index;
	float sum = FLT_MIN;
	if (idx != ntile)
	{
        #pragma unroll
		inner_loop(WORKSIZE_n);
	}
	else
	{
        #pragma unroll 4
		inner_loop2(n - idx - tid);
	}
	__shared__ volatile float smem[BLOCKSIZE_n];
	__shared__ volatile int imem[BLOCKSIZE_n];
	block_reduce_max<BLOCKSIZE_n>(x, y, smem, imem, sum, index, tid);
}

