#include "atomicOperations.h"
#include "snrm2_macros.h"

// 
// Generic Euclidian norm kernel.
//
// - Designed for architectures with atomicCAS, i.e., sm_11 and higher.
// - Simple design for all capabilities
// - Handled by splitting matrix as
// --------------------------
// |                 |      |
// |                 |      |
// |                 |      |
// | left tiled part | rest | 
// |                 |      |
// |                 |      |
// |                 |      |
// --------------------------
// - Kernel is generic and works for all cases
// - Unrolling of inner loops
//     1) to reduce instruction count from loop
//     2) to collect shared loads into 128 bit transactions (compiler should do that)
//   * Level determined by WORKSIZE_n
//   * For WORKSIZE_n = 1:
//       Given by a constant: full unroll
//       Given by a variable: unroll level 4
__device__ unsigned int count = 0; 
__shared__ bool isLastBlockDone;

template <int BLOCKSIZE_n>
inline __device__ void block_reduce(float *y, volatile float *smem, float sum, int tid)
{
	smem[tid] = sum;
	__syncthreads();
	if (BLOCKSIZE_n >= 1024) { if (tid < 512) { smem[tid] += smem[tid + 512]; } __syncthreads(); }
	if (BLOCKSIZE_n >= 512) { if (tid < 256) { smem[tid] += smem[tid + 256]; } __syncthreads(); }
	if (BLOCKSIZE_n == 480) { if (tid < 224) { smem[tid] += smem[tid + 256]; } __syncthreads(); }
	if (BLOCKSIZE_n == 448) { if (tid < 192) { smem[tid] += smem[tid + 256]; } __syncthreads(); }
	if (BLOCKSIZE_n == 416) { if (tid < 160) { smem[tid] += smem[tid + 256]; } __syncthreads(); }
	if (BLOCKSIZE_n == 384) { if (tid < 128) { smem[tid] += smem[tid + 256]; } __syncthreads(); }
	if (BLOCKSIZE_n >= 256) { if (tid < 128) { smem[tid] += smem[tid + 128]; } __syncthreads(); }
	if (BLOCKSIZE_n == 352) { if (tid < 96) { smem[tid] += smem[tid + 256]; } __syncthreads(); }
	if (BLOCKSIZE_n == 224) { if (tid < 96) { smem[tid] += smem[tid + 128]; } __syncthreads(); }
	if (BLOCKSIZE_n == 320) { if (tid < 64) { smem[tid] += smem[tid + 256]; } __syncthreads(); }
	if (BLOCKSIZE_n == 192) { if (tid < 64) { smem[tid] += smem[tid + 128]; } __syncthreads(); }
	if (BLOCKSIZE_n >= 128) { if (tid < 64) { smem[tid] += smem[tid + 64]; } __syncthreads(); }
	if (BLOCKSIZE_n == 288) { if (tid < 32) { smem[tid] += smem[tid + 256]; } __syncthreads(); }
	if (BLOCKSIZE_n == 160) { if (tid < 32) { smem[tid] += smem[tid + 128]; } __syncthreads(); }
	if (BLOCKSIZE_n == 96) { if (tid < 32) { smem[tid] += smem[tid + 64]; } __syncthreads(); }
	if (BLOCKSIZE_n >= 64) { if (tid < 32) { smem[tid] += smem[tid + 32]; } }
	if (BLOCKSIZE_n >= 32) { if (tid < 16) { smem[tid] += smem[tid + 16]; } }
	if (BLOCKSIZE_n >= 16) { if (tid < 8) { smem[tid] += smem[tid + 8]; } }
	if (BLOCKSIZE_n >= 8) { if (tid < 4) { smem[tid] += smem[tid + 4]; } }
	if (BLOCKSIZE_n >= 4) { if (tid < 2) { smem[tid] += smem[tid + 2]; } }
	if (tid == 0) atomicAdd(y, smem[tid] + smem[tid + 1]); 
}

template <int BLOCKSIZE_n, int WORKSIZE_m, int WORKSIZE_n>
__launch_bounds__(BLOCKSIZE_n, MIN_BLOCKS_PER_MP)
__global__ void snrm2_kernel(const float *x, float *y, int n, int ntile)
{
	const int tid = threadIdx.x;
	const int idx = blockIdx.x * BLOCKSIZE_n * WORKSIZE_n;
	x += idx + tid;
	float sum = 0.0;
	if (idx != ntile)
	{
        #pragma unroll
		inner_loop(WORKSIZE_n);
	}
	else
	{
        #pragma unroll 4
		inner_loop2(n - idx - tid);
	}
	__shared__ volatile float smem[BLOCKSIZE_n];
	block_reduce<BLOCKSIZE_n>(y, smem, sum, tid);
	if (tid == 0) 
	{
		unsigned int value = atomicInc(&count, gridDim.x); 
		isLastBlockDone = (value == (gridDim.x - 1)); 
	}
	__syncthreads(); 
	if (isLastBlockDone) {
		y[0] = sqrt(y[0]);
		count = 0;
	}
}

template <int BLOCKSIZE_n, int WORKSIZE_m, int WORKSIZE_n>
__launch_bounds__(BLOCKSIZE_n, MIN_BLOCKS_PER_MP)
__global__ void snrm2_kernel2(const float* x, float* y, int n, int ntile)
{
	const int tid = threadIdx.x;
	const int idx = blockIdx.x * BLOCKSIZE_n * WORKSIZE_n;
	x += idx + tid;
	float sum = 0.0;
	if (idx != ntile)
	{
		if (WORKSIZE_m == 1)
		{	// NB: full unroll
            #pragma unroll
			inner_loop(WORKSIZE_n);
		}
		if (WORKSIZE_m == 2)
		{
            #pragma unroll 2
			inner_loop(WORKSIZE_n);
		}
		if (WORKSIZE_m == 3)
		{
            #pragma unroll 3
			inner_loop(WORKSIZE_n);
		}
		if (WORKSIZE_m == 4)
		{
            #pragma unroll 4
			inner_loop(WORKSIZE_n);
		}
		if (WORKSIZE_m == 5)
		{
            #pragma unroll 5 
			inner_loop(WORKSIZE_n);
		}
		if (WORKSIZE_m == 6)
		{
            #pragma unroll 6
			inner_loop(WORKSIZE_n);
		}
		if (WORKSIZE_m == 7)
		{
            #pragma unroll 7
			inner_loop(WORKSIZE_n);
		}
		if (WORKSIZE_m == 8)
		{
            #pragma unroll 8
			inner_loop(WORKSIZE_n);
		}
	}
	else
	{
		if (WORKSIZE_m == 1)
		{
            #pragma unroll 4
			inner_loop2(n - idx - tid);
		}
		if (WORKSIZE_m == 2)
		{
            #pragma unroll 2
			inner_loop2(n - idx - tid);
		}
		if (WORKSIZE_m == 3)
		{
            #pragma unroll 3
			inner_loop2(n - idx - tid);
		}
		if (WORKSIZE_m == 4)
		{
            #pragma unroll 4
			inner_loop2(n - idx - tid);
		}
		if (WORKSIZE_m == 5)
		{
            #pragma unroll 5 
			inner_loop2(n - idx - tid);
		}
		if (WORKSIZE_m == 6)
		{
            #pragma unroll 6
			inner_loop2(n - idx - tid);
		}
		if (WORKSIZE_m == 7)
		{
            #pragma unroll 7
			inner_loop2(n - idx - tid);
		}
		if (WORKSIZE_m == 8)
		{
            #pragma unroll 8
			inner_loop2(n - idx - tid);
		}
	}
	__shared__ volatile float smem[BLOCKSIZE_n];
	block_reduce<BLOCKSIZE_n>(y, smem, sum, tid);
	if (tid == 0) 
	{
		unsigned int value = atomicInc(&count, gridDim.x); 
		isLastBlockDone = (value == (gridDim.x - 1)); 
	}
	__syncthreads(); 
	if (isLastBlockDone) {
		y[0] = sqrt(y[0]);
		count = 0;
	}
}

